% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sits_active_learning.R
\name{sits_uncertainty_sampling}
\alias{sits_uncertainty_sampling}
\title{Suggest samples for enhancing classification accuracy}
\usage{
sits_uncertainty_sampling(cube, n = 100, min_dist_pixels = 10)
}
\arguments{
\item{cube}{A `sits` uncertainty cube. See `sits_uncertainty`.}

\item{n}{Number of suggested points.}

\item{min_dist_pixels}{Minimum distance among suggested points (in pixels).}
}
\value{
A data.frame with longitude & latitude in WGS84 with locations
            which have high uncertainty and meet the minimum distance
            criteria.
}
\description{
Suggest samples for regions of high uncertainty as predicted by the model.
The function selects data points that have confused an algorithm.
These points don't have labels and need be manually labelled by experts
and then used to increase the classification's training set.

This function is best used in the following context
\itemize{
   \item{1. }{Select an initial set of samples.}
   \item{2. }{Train a machine learning model.}
   \item{3. }{Build a data cube and classify it using the model.}
   \item{4. }{Run a Bayesian smoothing in the resulting probability cube.}
   \item{5. }{Create an uncertainty cube.}
   \item{6. }{Perform uncertainty sampling.}
}
The Bayesian smoothing procedure will reduce the classification outliers
and thus increase the likelihood that the resulting pixels with high
uncertainty have meaningful information.
}
\examples{
if (sits_run_examples()) {
    # create a data cube
    data_dir <- system.file("extdata/raster/mod13q1", package = "sits")
    cube <- sits_cube(
        source = "BDC",
        collection = "MOD13Q1-6",
        data_dir = data_dir,
        delim = "_",
        parse_info = c("X1", "X2", "tile", "band", "date")
    )
    # build a random forest model
    samples_ndvi <- sits_select(samples_modis_4bands, bands = c("NDVI"))
    rfor_model <- sits_train(samples_ndvi, ml_method = sits_rfor())
    # classify the cube
    probs_cube <- sits_classify(data = cube, ml_model = rfor_model)
    # run a Bayesian smoothing
    bayes_cube <- sits_smooth(probs_cube)
    # create an uncertainty cube
    uncert_cube <- sits_uncertainty(bayes_cube)
    # obtain a new set of samples for active learning
    # the samples are located in uncertain places
    new_samples <- sits_uncertainty_sampling(uncert_cube)
}

}
\references{
Robert Monarch, "Human-in-the-Loop Machine Learning: Active learning
and annotation for human-centered AI". Manning Publications, 2021.
}
\author{
Alber Sanchez, \email{alber.sanchez@inpe.br}

Rolf Simoes, \email{rolf.simoes@inpe.br}

Felipe Carvalho, \email{felipe.carvalho@inpe.br}

Gilberto Camara, \email{gilberto.camara@inpe.br}
}
